/*
    This utility function assembles the matrices for a finite 
    element program which does analysis on a quad.  It is for
    eletromagnetism.

    This code solves for the more limited problem of finding
    the cut-off frequencies and corresponding eigenmodes.
    Unlike qdEKassemble which assembles all the matrices to solve
    for the system given by Equation 8.49 on page 247 of

        Jin, Jianming, The Finite Element Method in Electromagnetics,
            John Wiley & Sons, New York, 1993.

    I only assemble for the matrices given by 8.41 and solve for ko.
    This means that Att is given by the first term in the integral
    of 8.49 and Btt is the second.

		Updated 12/16/09

    SLFCEM source file
    Version:  1.1
    Copyright (C) 1999-2009  San Le 

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.
 
*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "qdeconst.h"
#include "qdestruct.h"

extern int analysis_flag, numed, numel, numnp, sof, modal_flag, static_flag;
extern int AB_matrix_store, Bzz_matrix_store;
extern int LU_decomp_flag, Bzz_LU_decomp_flag, numel_EM, numel_EM_K, numel_EM_P;
extern double dcdx[nsdsq*num_int], shg[sosh], shg_node[sosh], shl[sosh],
	shg_E[sosh_E], shg_node_E[sosh_E], curl_E[num_int*epel], shl_node[sosh],
	shl_node2[sosh_node2], w[num_int], *Area0, sigma;

int globalConjKassemble(double *, int *, int , double *,
	double *, int , int , int );

int globalKassemble(double *, int *, double *, int *, int );

int matX( double *,double *,double *, int ,int ,int );

int dyadicX( double *,double *,double *, int);

int matXT( double *, double *, double *, int, int, int);

int quadB_E( double *, double *, double *, double *, double *, double *, double *,
	double *, double *, double *);

int qdshg_E( double *, double *, int, double *, double *, double *, double *,
	double *, double *, double *);

int qdEKassemble(double *Att, double *Att_sigma, double *Btt, double *Btz, double *Bzt,
	double *Bzz, int *connect, int *node_edge_connect, int *edge_connect,
	double *coord, int *el_matl, int *id, int *idiag, double *Att_diag,
	double *Att_sigma_diag, double *Bzz_diag, int *lm, MATL *matl, double *node_counter,
	double *edge_counter)
{
	int i, i1, i2, j, k, dof_el[neqel], EMdof_el[neqel4], sdof_el[npel*nsd];
	int check, counter, node, node0, node1, node2, node3, edge;
	int matl_num;
	double permit, permea, kosq, G;
	double B[npel], gradB[soB8], BXB[neqlsq16], Bedge[soB8], curlBedge[epel],
		Bedge2[soB8], curlBedge2[epel];
	double Att_temp[neqlsq16], Btt_temp[neqlsq16];
	double Att_el[neqlsq16], Att_sigma_el[neqlsq16], Btt_el[neqlsq16];
	double coord_el_trans[npel*nsd], length[epel], lengthsq[epel];
	double det[num_int];
	double fdum, fdum2;
	double shg_sign[4];

	for( k = 0; k < numel; ++k )
	{

		matl_num = *(el_matl+k);
		permit = matl[matl_num].er;
		permea = matl[matl_num].ur;
		kosq = matl[matl_num].ko*matl[matl_num].ko;

		/*printf("lamda, mu, permit, permea  %f %f %f %f \n", lamda, mu, Emod, Pois);*/

/* Create the coord_el transpose vector for one element */

		for( j = 0; j < npel; ++j )
		{
			node = *(connect+npel*k+j);

			*(sdof_el+nsd*j) = nsd*node;
			*(sdof_el+nsd*j+1) = nsd*node+1;

			*(coord_el_trans+j) = *(coord+*(sdof_el+nsd*j));
			*(coord_el_trans+npel*1+j) = *(coord+*(sdof_el+nsd*j+1));

			*(dof_el+ndof*j) = ndof*node;
			*(dof_el+ndof*j+1) = ndof*node+1;

/* Count the number of times a particular node is part of an element */

			if(analysis_flag == 1)
				*(node_counter + node) += 1.0;
		}

/* Create edge_el for one element */

		for( j = 0; j < epel; ++j )
		{
			edge = *(edge_connect+epel*k+j);
			node1 = *(node_edge_connect+edge*nped);
			node2 = *(node_edge_connect+edge*nped + 1);

			*(EMdof_el+edof*j) = edof*edge;

/* Calculate the edge lengths */

			fdum = *(coord+node2*nsd) - *(coord+node1*nsd);
			fdum2 = *(coord+node2*nsd+1) - *(coord+node1*nsd+1);
			*(lengthsq + j) = fdum*fdum + fdum2*fdum2;
			*(length + j) = sqrt(*(lengthsq + j));

/* Count the number of times a particular edge is part of an element */

			if(analysis_flag == 1)
				*(edge_counter + edge) += 1.0;
		}

/* The variables below are needed to address the issue explained in:

   ~/slfcem-1.1/EDGE-ELEMENTS

   Unlike the triangle element, the exact way in which a negative should
   be assigned to an edge shape function must be done carefully, which is
   why you get the odd numbering below.  If you look in Jin's book, on
   figure 8.4a on page 238, you will see that the edges are numbered in
   a particular way and that there is an assumed direction for the
   electric field along an edge.  This must be adhered to when assigning
   "shg_sign" below.  For instance, you can reverse the inequality sign
   on 2 opposing edges like 1 and 2 or 3 and 4.  It may be due to the fact
   that all elements must have an equal number of clockwise and counter-clocwise
   directions for the electic field on an edge.  I did some test using the
   file "rect3a" in terms of shifting the node numbering, which you can read
   about at the bottom of that file.
*/

		node0 = *(connect+npel*k);
		node1 = *(connect+npel*k+1);
		node2 = *(connect+npel*k+2);
		node3 = *(connect+npel*k+3);

		*(shg_sign)   = 1.0; *(shg_sign+1) = 1.0;
		*(shg_sign+2) = 1.0; *(shg_sign+3) = 1.0;

		if(node0 > node1) *(shg_sign)   = -1.0;
		if(node3 > node2) *(shg_sign+1) = -1.0;
		if(node0 > node3) *(shg_sign+2) = -1.0;
		if(node1 > node2) *(shg_sign+3) = -1.0;

/* Assembly of the dcdx and shg matrix for each integration point */

		check=qdshg_E(dcdx, det, k, shl, shg, shg_E, curl_E,
			coord_el_trans, length, shg_sign);
		if(!check) printf( "Problems with qdshg_E\n");

/* The loop over j below calculates the 4 points of the gaussian integration 
   for several quantities */

		memset(Att_el,0,neqlsq16*sof);
		memset(Att_sigma_el,0,neqlsq16*sof);
		memset(Btt_el,0,neqlsq16*sof);

		for( j = 0; j < num_int; ++j )
		{

		    memset(B,0,npel*sof);
		    memset(gradB,0,soB8*sof);
		    memset(Bedge,0,soB8*sof);
		    memset(curlBedge,0,epel*sof);
		    memset(Att_temp,0,neqlsq16*sof);
		    memset(Btt_temp,0,neqlsq16*sof);
		    memset(BXB,0,neqlsq16*sof);

/* Assembly of the component matrices */

		    check = quadB_E((dcdx+nsdsq*j), (shl+npel*(nsd+1)*j),
			(shg+npel*(nsd+1)*j), (shg_E+epel*nsd2*j), (curl_E+epel*j),
			B, gradB, Bedge, curlBedge, length);
		    if(!check) printf( "Problems with quadB_E\n");

		    memcpy(curlBedge2, curlBedge, epel*sizeof(double));
		    memcpy(Bedge2, Bedge, soB8*sizeof(double));

		    check=dyadicX(Att_temp, curlBedge, curlBedge2, neqel4);
		    if(!check) printf( "Problems with dyadicX \n");

		    check=matXT(Btt_temp, Bedge, Bedge2, neqel4, neqel4, nsd);
		    if(!check) printf( "Problems with matXT  \n");

		    fdum = *(w+j)*(*(det+j));

		    for( i2 = 0; i2 < neqlsq16; ++i2 )
		    {
			  *(Btt_el+i2) += *(Btt_temp+i2)*fdum*permit;
			  *(Att_el+i2) += *(Att_temp+i2)*fdum/permea;
		    }
		}

/* Att_sigma is the modified Att matrix which makes the eigenvalue calculation
   find the eigenvalue nearest the value of "sigma".  See page 583-584 of:

      Hughes, Thomas J. R., The Finite Element Method, Prentice-Hall Inc.,
          Englewood Cliffs,  1987.

*/
		for( i2 = 0; i2 < neqlsq16; ++i2 )
		{
		    *(Att_sigma_el + i2) = *(Att_el + i2) - sigma*(*(Btt_el + i2));
		}


/*
		printf("\n Att_el %4d\n", k);

		for( i2 = 0; i2 < neqel4; ++i2 )
		{
			printf("    ");
			for( i1 = 0; i1 < neqel4; ++i1 )
			{
				printf("%12.8f ",*(Att_sigma_el + neqel4*i2 + i1));
			}
			printf("    %12.8f \n",*(length + i2));
		}
		printf("\n");
*/


/*
		printf("\n Btt_el %4d\n", k);

		for( i2 = 0; i2 < neqel4; ++i2 )
		{
			printf("    ");
			for( i1 = 0; i1 < neqel4; ++i1 )
			{
				printf("%12.8f ",*(Btt_el + neqel4*i2 + i1));
			}
			printf("    %12.8f \n",*(length + i2));
		}
		printf("\n");
*/



		if(analysis_flag == 1)
		{

#if 1
/* Assembly of either the global skylined Att matrix or numel_EM of the
   element Att matrices if the Conjugate Gradient method is used */

			if(LU_decomp_flag)
			{
			    /*check = globalKassemble(Att, idiag, Att_el, (lm + k*neqel4),
				neqel4);
			    if(!check) printf( "Problems with globalKassemble \n");*/

			    check = globalKassemble(Att_sigma, idiag, Att_sigma_el,
				(lm + k*neqel4), neqel4);
			    if(!check) printf( "Problems with globalKassemble \n");
			}
			else
			{
			    /*check = globalConjKassemble(Att, EMdof_el, k, Att_diag, Att_el,
				neqel4, neqlsq16, numel_EM_K);
			    if(!check) printf( "Problems with globalConjKassemble \n");*/

			    check = globalConjKassemble(Att_sigma, EMdof_el, k,
				Att_sigma_diag, Att_sigma_el, neqel4, neqlsq16, numel_EM_K);
			    if(!check) printf( "Problems with globalConjKassemble \n");
			}
#endif

/* If there is enough memory, store all the element Att and Btt matrices */

			if(AB_matrix_store)
			{
			   for( j = 0; j < neqlsq16; ++j )
			   {
				*(Btt + neqlsq16*k + j) = *(Btt_el + j);
			   }

			   for( j = 0; j < neqlsq16; ++j )
			   {
				*(Att + neqlsq16*k + j) = *(Att_el + j);
			   }

			}

		}
	}


	return 1;
}

