/*
    This utility function calculates the value of the electric field for each node
    on a 4 node quad element.  The first function, "qdEMnode_calculate" does this
    for the edge based elements used in the calculation of waveguides.  The second,
    "qdEVnode_calculate" is for electrostatic problems.

		Updated 12/18/09

    SLFCEM source file
    Version:  1.1
    Copyright (C) 2009  San Le 

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.
 
*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "qdeconst.h"
#include "qdestruct.h"

extern int dof, numnp, numel;
extern double dcdx[nsdsq*num_int], shg[sosh], shg_node[sosh], shl[sosh],
        shg_E[sosh_E], shg_node_E[sosh_E], curl_E[num_int*epel], shl_node[sosh],
	shl_node2[sosh_node2], w[num_int], *Area0;

int matXT(double *, double *, double *, int, int, int);

int qdshg_E( double *, double *, int , double *, double *, double *, double *,
	double *, double *, double *);

int qdEMnode_calculate(int *connect, int *edge_connect, int *node_edge_connect,
	double *coord, double *EMedge, double *EMnode, double *EMmag,
	double *node_counter)
{
/*  The electric field is calculated at all four nodes of an element based on
    interpolating the calculated electric field edge values.  Each edge
    has a tangentinal and normal component which will be decomposed
    into X and Y global directions at each node.  Because there may be
    multiple edges which contain the node, averaging is done at the
    end based on the number of edges which have the node.  This is used
    when doing the waveguide problem.

		Updated 12/15/09
*/

	int i, i1, i2, j, k, dof_el[neqel], EMdof_el[neqel4], sdof_el[npel*nsd];;
	int check, counter, node, node0, node1, node2, node3, node_el[npel];
	int matl_num, edge;
	double fdum, fdum1, fdum2, fdum3, fdum4;
	double EMedge_el[neqel4], EMnode_el[nsd];
	double coord_el_trans[npel*nsd], length[epel], lengthsq[epel];
	double Bedge[soB8];
	double det[num_int];
	double shg_sign[4];

/* Note below that I am still looping through the elements rather than the number of
   edges, "numed".  The reason is because I have to use the edge shape functions
   to get the normal electric field components to an edge.  These shape functions need the
   edge connectivity that is based on the elements.  Also, note that the edge numbering
   is not counter-clockwise.  Look at Figure 8.4a on page 238 of:

      Jin, Jianming, The Finite Element Method in Electromagnetics, John Wiley & Sons,
         New York, 1993.

   Basically, the directions of the electric field and edge numbering are:

                -> 1
              _______
           /\ |      | /\
           |  |      | |
            2 |      | 3
              |------| 
                 -> 0
*/

	memset(EMnode,0,dof*sizeof(double));
	memset(EMmag,0,numnp*sizeof(double));

	for( k = 0; k < numel; ++k )
	{

/* Create the coord transpose vector and other variables for one element */

	    for( j = 0; j < npel; ++j )
	    {
		*(node_el+j) = node = *(connect+npel*k+j);

		*(sdof_el+nsd*j)=nsd*node;
		*(sdof_el+nsd*j+1)=nsd*node+1;

		*(coord_el_trans+j) =
			*(coord+*(sdof_el+nsd*j));
		*(coord_el_trans+npel*1+j) =
			*(coord+*(sdof_el+nsd*j+1));

		*(dof_el+ndof*j) = ndof*node;
		*(dof_el+ndof*j+1) = ndof*node+1;

	    }

/* Create edge_el for one element */

	    for( j = 0; j < epel; ++j )
	    {
		    edge = *(edge_connect+epel*k+j);
		    node1 = *(node_edge_connect+edge*nped);
		    node2 = *(node_edge_connect+edge*nped + 1);

		    *(EMdof_el+edof*j) = edof*edge;

/* Calculate the edge lengths */

		    fdum = *(coord+node2*nsd) - *(coord+node1*nsd);
		    fdum2 = *(coord+node2*nsd+1) - *(coord+node1*nsd+1);
		    *(lengthsq + j) = fdum*fdum + fdum2*fdum2;
		    *(length + j) = sqrt(*(lengthsq + j));

	    }

	    for( j = 0; j < neqel4; ++j )
	    {
		*(EMedge_el + j) = *(EMedge + *(EMdof_el+j));
		/*printf("%12.8e  %4d\n", *(EMedge_el + j), *(EMdof_el+j));*/
	    }

/* The variables below are needed to address the issue explained in:

   ~/slfcem-1.1/EDGE-ELEMENTS

   Read in the code "qdekasmbl.c" to get more discussion of how I set
   "shg_sign".
*/

	    node0 = *(connect+npel*k);
	    node1 = *(connect+npel*k+1);
	    node2 = *(connect+npel*k+2);
	    node3 = *(connect+npel*k+3);

	    *(shg_sign)   = 1.0; *(shg_sign+1) = 1.0;
	    *(shg_sign+2) = 1.0; *(shg_sign+3) = 1.0;

	    if(node0 > node1) *(shg_sign)   = -1.0;
	    if(node3 > node2) *(shg_sign+1) = -1.0;
	    if(node0 > node3) *(shg_sign+2) = -1.0;
	    if(node1 > node2) *(shg_sign+3) = -1.0;

/* Assembly of the dcdx and shg matrix for each nodal point */

	    check=qdshg_E(dcdx, det, k, shl_node, shg_node, shg_node_E, curl_E,
		coord_el_trans, length, shg_sign);
	    if(!check) printf( "Problems with qdshg_E\n");

	    for( j = 0; j < num_int; ++j )
	    {
		*(Bedge)            = *(shg_node_E + epel*nsd2*j);
		*(Bedge + 1)        = *(shg_node_E + epel*nsd2*j + 1);
		*(Bedge + 2)        = *(shg_node_E + epel*nsd2*j + 2);
		*(Bedge + 3)        = *(shg_node_E + epel*nsd2*j + 3);
		*(Bedge+neqel4*1)   = *(shg_node_E + epel*nsd2*j + epel*1);
		*(Bedge+neqel4*1+1) = *(shg_node_E + epel*nsd2*j + epel*1 + 1);
		*(Bedge+neqel4*1+2) = *(shg_node_E + epel*nsd2*j + epel*1 + 2);
		*(Bedge+neqel4*1+3) = *(shg_node_E + epel*nsd2*j + epel*1 + 3);

		check = matX(EMnode_el, Bedge, EMedge_el, nsd2, 1, neqel4);
		if(!check) printf( "Problems with matX \n");

/* Note that the EMnode_el is not like the U_el in my mechanics codes.  This EMnode_el[2]
   is the x and y component of the electric field at node_el[j].   And an additional
   fact is that all intepolation is to a point, even if it edges are used as the
   interpolating quantity.

   The interpolation of EMnode is based on equation 8.38 on page 246 of Jin.

   The electric field magnitude, "EMmag", must be calculated as shown because
   at the location of the point charge, the different components will cancel
   each other out, but this is where the field is strongest.
*/
		*(EMnode + ndof*(*(node_el+j)))     += *(EMnode_el);
		*(EMnode + ndof*(*(node_el+j)) + 1) += *(EMnode_el + 1);

		fdum  = *(EMnode_el);
		fdum2 = *(EMnode_el+1);
		*(EMmag + *(node_el+j)) += sqrt(fdum*fdum + fdum2*fdum2);

		/*printf("%12.8e %12.8e %12.8e %12.8e\n",
			*(Bedge), *(Bedge+1), *(Bedge+2), *(Bedge+3)); 
		printf("%12.8e %12.8e %12.8e %12.8e\n",
			*(Bedge+neqel4), *(Bedge+neqel4+1), *(Bedge+neqel4+2), *(Bedge+neqel4+3)); 

		printf("EMnode %12.8e %12.8e\n", *(EMnode_el), *(EMnode_el + 1)); */
	    }

	}

/* Average all the components of EMnode depending based on its node count */

	for( i = 0; i < numnp ; ++i )
	{
		*(EMnode + ndof*i)   /= *(node_counter+i);
		*(EMnode + ndof*i+1) /= *(node_counter+i);
		*(EMmag + i) /= *(node_counter+i);
	}

	return 1;
}

int qdEVnode_calculate(int *connect, double *coord, double *Volt, double *EMnode, double *EMmag,
	double *node_counter)
{
/*  The electric field is calculated at all four nodes of an element based on
    the gradient of the voltage potential.  Because there may be multiple
    elements which contain the node, averaging is done at the end based on
    the number of times the node appears.  This function is used when doing the
    electrostatic problem.

		Updated 12/18/09
*/
	int i, i1, i2, j, k, dof_el[neqel], Vdof_el[neqel4], sdof_el[npel*nsd];;
	int check, counter, node, node0, node1, node2, node3, node_el[npel];
	int matl_num, edge;
	double fdum, fdum1, fdum2, fdum3, fdum4;
	double gradB[soB8];
	double Volt_el[Vneqel4], EMnode_el[nsd];
	double coord_el_trans[npel*nsd];
	double det[num_int];

	memset(EMnode,0,dof*sizeof(double));
	memset(EMmag,0,numnp*sizeof(double));

	for( k = 0; k < numel; ++k )
	{

/* Create the coord transpose vector and other variables for one element */

	    for( j = 0; j < npel; ++j )
	    {
		    *(node_el+j) = node = *(connect+npel*k+j);

		    *(sdof_el+nsd*j) = nsd*node;
		    *(sdof_el+nsd*j+1) = nsd*node+1;

		    *(coord_el_trans+j) = *(coord+*(sdof_el+nsd*j));
		    *(coord_el_trans+npel*1+j) = *(coord+*(sdof_el+nsd*j+1));

		    *(dof_el+ndof*j) = ndof*node;
		    *(dof_el+ndof*j+1) = ndof*node+1;

		    *(Vdof_el+Vndof*j) = Vndof*node;
	    }

	    for( j = 0; j < Vneqel4; ++j )
	    {
		*(Volt_el + j) = *(Volt + *(Vdof_el+j));
		/*printf("%12.8e  %4d\n", *(Volt_el + j), *(Vdof_el+j));*/
	    }

	    check=qdshg( det, k, shl_node, shg_node, coord_el_trans);
	    if(!check) printf( "Problems with qdshg\n");

	    /*printf("element %4d\n", k);*/
	    for( j = 0; j < npel; ++j )
	    {
		*(gradB)              = *(shg_node + npel*(nsd2+1)*j);
		*(gradB + 1)          = *(shg_node + npel*(nsd2+1)*j + 1);
		*(gradB + 2)          = *(shg_node + npel*(nsd2+1)*j + 2);
		*(gradB + 3)          = *(shg_node + npel*(nsd2+1)*j + 3);

		*(gradB + npel*1)     = *(shg_node + npel*(nsd2+1)*j + npel*1);
		*(gradB + npel*1 + 1) = *(shg_node + npel*(nsd2+1)*j + npel*1 + 1);
		*(gradB + npel*1 + 2) = *(shg_node + npel*(nsd2+1)*j + npel*1 + 2);
		*(gradB + npel*1 + 3) = *(shg_node + npel*(nsd2+1)*j + npel*1 + 3);

		check = matX(EMnode_el, gradB, Volt_el, nsd2, 1, Vneqel4);
		if(!check) printf( "Problems with matX \n");

/* Note that the EMnode_el is not like the U_el in my mechanics codes.  This EMnode_el[2]
   is the x and y component of the electric field at node_el[j].

   The electric field is the negative of the gradient of the potential.

   The electric field magnitude, "EMmag", must be calculated as shown because
   at the location of the point charge, the different components will cancel
   each other out, but this is where the field is strongest.
*/
		*(EMnode + ndof*(*(node_el+j)))     -= *(EMnode_el);
		*(EMnode + ndof*(*(node_el+j)) + 1) -= *(EMnode_el + 1);

		fdum  = *(EMnode_el);
		fdum2 = *(EMnode_el+1);
		*(EMmag + *(node_el+j)) += sqrt(fdum*fdum + fdum2*fdum2);

		/*printf("%16.8e %16.8e %16.8e %16.8e\n",
			*(gradB), *(gradB + 1), *(gradB + 2), *(gradB + 3));
		printf("%16.8e %16.8e %16.8e %16.8e\n",
			*(gradB + npel*1), *(gradB + npel*1 + 1), *(gradB + npel*1 + 2), *(gradB + npel*1 + 3));
		printf("Volt_el %16.8e %16.8e %16.8e %16.8e\n",
			*(Volt_el), *(Volt_el + 1), *(Volt_el + 2), *(Volt_el + 3));*/

		/*printf("EMnode %12.8e %12.8e\n", *(EMnode_el), *(EMnode_el + 1)); */
	    }

	}

/* Average all the components of EMnode depending based on its node count */

	for( i = 0; i < numnp ; ++i )
	{
		*(EMnode + ndof*i)   /= *(node_counter+i);
		*(EMnode + ndof*i+1) /= *(node_counter+i);
		*(EMmag + i) /= *(node_counter+i);
	}

	return 1;
}

