/*
     SLFCEM source file
     Version:  1.1
     Copyright (C) 1999-2009  San Le

     The source code contained in this file is released under the
     terms of the GNU Library General Public License.

     This file is almost identical to:

        ~/slffea-1.6/quad/quad/qdshape.c

     The differences are that I have to calculate:

        dc/dx, shg_E, curl_E

     in "qdshg_E" which is needed for the shape functions in
     the edge based interpolations. 

     I also calculate the element area using the determinents.
     The file:

        ~/slffea-1.6/quad/quad/qdarea.c

     is for 3-D quad elements so it uses a much more complex
     calculation.
*/

#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "qdeconst.h"

int dotX(double *,double *, double *, int );

int qdshl( double g, double *shl, double *w)
{
/* 
     This subroutine calculates the local shape function derivatives for
     a quad element at the gauss points.

     It is a C translation of the subroutine QDCSHL from the book
     "The Finite Element Method" by Thomas Hughes, page 784.

                          Updated 4/15/99
*/

	double ra[]={-0.50, 0.50, 0.50,-0.50};
	double sa[]={-0.50,-0.50, 0.50, 0.50};
	double temps,tempr,r,s;
	int i,j,k;

	for( k = 0; k < num_int; ++k )
	{
/* calculating the weights and local dN/ds,dr matrix */

		*(w+k)=1.0;

		r=g*(*(ra+k));
		s=g*(*(sa+k));
		for( i = 0; i < npel; ++i )
		{
		   tempr = pt5 + *(ra+i)*r;
		   temps = pt5 + *(sa+i)*s;
		   *(shl+npel*(nsd2+1)*k+i)=*(ra+i)*temps;
		   *(shl+npel*(nsd2+1)*k+npel*1+i)=tempr*(*(sa+i));
		   *(shl+npel*(nsd2+1)*k+npel*2+i)=tempr*temps;
		}
		/*printf("\n");*/
	}
	return 1;
}


int qdshl_node2(double *shl_node2)
{
/* 
   This subroutine is a condensed version of the shape function shl_node.  I have
   removed all of the zero terms which made up 75% of all the values in shl_node.

                          Updated 4/16/99
*/

/* For node 0 */
	*(shl_node2) =   -0.50; *(shl_node2+1) =  0.50; *(shl_node2+2) = -0.50;
	*(shl_node2+3) =  0.50;

/* For node 1 */
	*(shl_node2+4) = -0.50; *(shl_node2+5)  =  0.50; *(shl_node2+6)  = -0.50;
	*(shl_node2+7) =  0.50;

/* For node 2 */
	*(shl_node2+8) =  0.50; *(shl_node2+9) = -0.50; *(shl_node2+10) = -0.50;
	*(shl_node2+11) =  0.50;

/* For node 3 */
	*(shl_node2+12) =  0.50; *(shl_node2+13) = -0.50; *(shl_node2+14) = -0.50;
	*(shl_node2+15) =  0.50;

	return 1;
}



int qdshg_E( double *dcdx, double *det, int el, double *shl, double *shg,
	double *shg_E, double *curl_E, double *xl, double *length, double *shg_sign)
{
/*
     This subroutine calculates the global shape function derivatives for
     an electromagnetic quad element at the gauss points.

     It is a C translation of the subroutine QDCSHG from the book
     "The Finite Element Method" by Thomas Hughes, page 783.

 ....  CALCULATE GLOBAL DERIVATIVES OF SHAPE FUNCTIONS AND
       JACOBIAN DETERMINANTS FOR A FOUR-NODE QUADRALATERAL ELEMENT

       *(xl+j+npel*i) = GLOBAL COORDINATES(I LOOPS OVER X,Y)
       *(det+k)  = JACOBIAN DETERMINANT
       *(dcdx+nsdsq*k)      = d(XI)/dx
       *(dcdx+nsdsq*k + 1)  = d(XI)/dy
       *(dcdx+nsdsq*k + 2)  = d(ETA)/dx
       *(dcdx+nsdsq*k + 3)  = d(ETA)/dy
       *(shl+npel*(nsd2+1)*k+i) = LOCAL ("XI") DERIVATIVE OF SHAPE FUNCTION
       *(shl+npel*(nsd2+1)*k+npel*1+i) = LOCAL ("ETA") DERIVATIVE OF SHAPE FUNCTION
       *(shl+npel*(nsd2+1)*k+npel*2+i) = LOCAL SHAPE FUNCTION
       *(shg+npel*(nsd2+1)*k+i) = X-DERIVATIVE OF SHAPE FUNCTION
       *(shg+npel*(nsd2+1)*k+npel*1+i) = Y-DERIVATIVE OF SHAPE FUNCTION
       *(shg+npel*(nsd2+1)*k+npel*2+i) = shl(npel*(nsd2+1)*k+npel*2+i)

       *(shg_E+npel*nsd2*k+npel*1+i) = LOCAL EDGE SHAPE FUNCTION

       *(curl_E+npel*nsd2*k+npel*1+i) = CURL OF LOCAL EDGE SHAPE FUNCTION

       *(xs+2*i+j) = JACOBIAN MATRIX
          i    = LOCAL NODE NUMBER OR GLOBAL COORDINATE NUMBER
          j    = GLOBAL COORDINATE NUMBER
          k    = INTEGRATION-POINT NUMBER
       num_int    = NUMBER OF INTEGRATION POINTS, EQ.8 

          i    = LOCAL EDGE NUMBER
       *(shg_E+epel*nsd2*k+i) = X COMPONENT OF SHAPE FUNCTION
       *(shg_E+epel*nsd2*k+epel*1+i) = Y COMPONENT OF SHAPE FUNCTION
       *(curl_E + i) = CURL OF SHAPE FUNCTION
	
                          Updated 12/12/09
*/

	double xs[4],temp;
	double dxidx, dxidy, detadx, detady;
	int check,i,j,k;

	memcpy(shg,shl,sosh*sizeof(double));

	for( k = 0; k < num_int; ++k )
	{

/* The jacobian dx/dc is calculated below */

	   for( j = 0; j < nsd2; ++j )
	   {
		for( i = 0; i < nsd2; ++i )
		{
		   check=dotX((xs+nsd2*i+j),(shg+npel*(nsd2+1)*k+npel*j),
			(xl+npel*i),npel);
		}
	   }

	   *(det+k)=*(xs)*(*(xs+3))-*(xs+2)*(*(xs+1));
	   /*printf(" %9.5f %9.5f\n %9.5f %9.5f\n",
		*(xs),*(xs+1),*(xs+nsd2*1),*(xs+nsd2*1+1));
	   printf("%d %f\n", k, *(det+k));*/

	   if(*(det+k) <= 0.0 ) 
	   {
		printf("the element (%d) is inverted; det:%f; integ pt.:%d\n",
			el,*(det+k),k);
		return 0;
	   }

/* The inverse of the jacobian, dc/dx, is calculated below and
   stored in dcdx
*/
	   temp=*(xs);
	   *(xs) = *(xs + 3)/(*(det+k));
	   *(xs + 1) *= -1./(*(det+k));
	   *(xs + 2) *= -1./(*(det+k));
	   *(xs + 3) = temp/(*(det+k));

	   *(dcdx + nsdsq*k) = *(xs);
	   *(dcdx + nsdsq*k + 1) = *(xs + 1);
	   *(dcdx + nsdsq*k + 2) = *(xs + 2);
	   *(dcdx + nsdsq*k + 3) = *(xs + 3);

	   for( i = 0; i < npel; ++i )
	   {
		*(shg+npel*(nsd2+1)*k+i) = *(xs)*(*(shl+npel*(nsd2+1)*k+i))+
			*(xs+2)*(*(shl+npel*(nsd2+1)*k+npel*1+i));
		*(shg+npel*(nsd2+1)*k+npel*1+i)=*(xs+1)*(*(shl+npel*(nsd2+1)*k+i))+
			*(xs+3)*(*(shl+npel*(nsd2+1)*k+npel*1+i));
		/*printf("%d %f %f %f\n", i, *(shg+npel*(nsd2+1)*k+i),
			*(shg+npel*(nsd2+1)*k+npel*1+i), 
			*(shg+npel*(nsd2+1)*k+npel*2+i));*/
	   }


/* 
     Calculate the local edge shape function derivatives for a quad element.

     Unfortunately, it is pretty much unreadable.  The only way to understand
     how it all works is to have all the notes I have taken.  What you need to do
     is compare the shape function derivatives for N, which I call "shg_E", with
     eq. 8.22-2.25 in Jin's book, page 239-240.  Also, compare it with my notes in
     the Black book (which unfortunately, only I have), for "shl" as well as my
     notes on the back of page 246 of Jin's book.  The curl of the shape functions,
     which are given by "curl_E" below, are not defined in the book.

     Note that the edge based interpolating shape functions below are different
     from the node based above.  This edge interpolation is not encountered
     in my other codes which is why it is expressed so ackwardly.

     Checked shg_E and curlBedge and they are correct.  The 1/4 is already
     in shl.

     In the breakdown of

        *(shg_E+epel*nsd2*k) = *(shg_sign)*(*(dxidx))*(*(shl+npel*(nsd+1)*k+1))*(*(length));

                                  + or -      dxi/dx               c1             length

           dxidx  =  d(xi) / d(x)
           detadx = d(eta) / d(x)
*/

	   dxidx  = *(dcdx + nsdsq*k);     dxidy  = *(dcdx + nsdsq*k + 1);
	   detadx = *(dcdx + nsdsq*k + 2); detady = *(dcdx + nsdsq*k + 3);

	   *(shg_E+epel*nsd2*k) =
		*(shg_sign)*dxidx*(*(shl+npel*(nsd+1)*k+1))*(*(length));
	   *(shg_E+epel*nsd2*k+1) =
		*(shg_sign+1)*dxidx*(*(shl+npel*(nsd+1)*k+2))*(*(length + 1));
	   *(shg_E+epel*nsd2*k+2) =
		*(shg_sign+2)*detadx*(*(shl+npel*(nsd+1)*k+npel*1+3))*(*(length + 2));
	   *(shg_E+epel*nsd2*k+3) =
		*(shg_sign+3)*detadx*(*(shl+npel*(nsd+1)*k+npel*1+2))*(*(length + 3));

	   *(shg_E+epel*nsd2*k+epel*1) =
		*(shg_sign)*dxidy*(*(shl+npel*(nsd+1)*k+1))*(*(length));
	   *(shg_E+epel*nsd2*k+epel*1+1) =
		*(shg_sign+1)*dxidy*(*(shl+npel*(nsd+1)*k+2))*(*(length + 1));
	   *(shg_E+epel*nsd2*k+epel*1+2) =
		*(shg_sign+2)*detady*(*(shl+npel*(nsd+1)*k+npel*1+3))*(*(length + 2));
	   *(shg_E+epel*nsd2*k+epel*1+3) =
		*(shg_sign+3)*detady*(*(shl+npel*(nsd+1)*k+npel*1+2))*(*(length + 3));

/* Calculate the curl of the shape functions above */

           *(curl_E + epel*k) =
                *(shg_sign)*(*(length))*(dxidx*detady-dxidy*detadx)/4.0;
           *(curl_E + epel*k + 1) =
                *(shg_sign+1)*(*(length + 1))*(detadx*dxidy-detady*dxidx)/4.0;
           *(curl_E + epel*k + 2) =
                *(shg_sign+2)*(*(length + 2))*(detadx*dxidy-detady*dxidx)/4.0;
           *(curl_E + epel*k + 3) =
                *(shg_sign+3)*(*(length + 3))*(dxidx*detady-dxidy*detadx)/4.0;

	    /*printf("\n");*/
	}
	return 1; 
}



int qdshg( double *det, int el, double *shl, double *shg, double *xl)
{
/*
     This subroutine calculates the global shape function derivatives for
     a quad element at the gauss points.

     It is a C translation of the subroutine QDCSHG from the book
     "The Finite Element Method" by Thomas Hughes, page 783.

 ....  CALCULATE GLOBAL DERIVATIVES OF SHAPE FUNCTIONS AND
       JACOBIAN DETERMINANTS FOR A FOUR-NODE QUADRALATERAL ELEMENT

       *(xl+j+npel*i) = GLOBAL COORDINATES(I LOOPS OVER X,Y)
       *(det+k)  = JACOBIAN DETERMINANT
       *(shl+npel*(nsd2+1)*k+i) = LOCAL ("XI") DERIVATIVE OF SHAPE FUNCTION
       *(shl+npel*(nsd2+1)*k+npel*1+i) = LOCAL ("ETA") DERIVATIVE OF SHAPE FUNCTION
       *(shl+npel*(nsd2+1)*k+npel*2+i) = LOCAL SHAPE FUNCTION
       *(shg+npel*(nsd2+1)*k+i) = X-DERIVATIVE OF SHAPE FUNCTION
       *(shg+npel*(nsd2+1)*k+npel*1+i) = Y-DERIVATIVE OF SHAPE FUNCTION
       *(shg+npel*(nsd2+1)*k+npel*2+i) = shl(npel*(nsd2+1)*k+npel*2+i)
       *(xs+2*i+j) = JACOBIAN MATRIX
          i    = LOCAL NODE NUMBER OR GLOBAL COORDINATE NUMBER
          j    = GLOBAL COORDINATE NUMBER
          k    = INTEGRATION-POINT NUMBER
       num_int    = NUMBER OF INTEGRATION POINTS, EQ.4 

                        Updated 8/7/06
*/

	double xs[4],temp;
	int check,i,j,k;

	memcpy(shg,shl,sosh*sizeof(double));

	for( k = 0; k < num_int; ++k )
	{

/* The jacobian dx/dc is calculated below */

	   for( j = 0; j < nsd2; ++j )
	   {
		for( i = 0; i < nsd2; ++i )
		{
		   check=dotX((xs+nsd2*i+j),(shg+npel*(nsd2+1)*k+npel*j),
			(xl+npel*i),npel);
		}
	   }

	   *(det+k)=*(xs)*(*(xs+3))-*(xs+2)*(*(xs+1));
	   /*printf(" %9.5f %9.5f\n %9.5f %9.5f\n",
		*(xs),*(xs+1),*(xs+nsd2*1),*(xs+nsd2*1+1));
	   printf("%d %f\n", k, *(det+k));*/

	   if(*(det+k) <= 0.0 ) 
	   {
		printf("the element (%d) is inverted; det:%f; integ pt.:%d\n",
			el,*(det+k),k);
		return 0;
	   }

/* The inverse of the jacobian, dc/dx, is calculated below */

	   temp=*(xs);
	   *(xs)=*(xs+3)/(*(det+k));
	   *(xs+1)*=-1./(*(det+k));
	   *(xs+2)*=-1./(*(det+k));
	   *(xs+3)=temp/(*(det+k));

	   for( i = 0; i < npel; ++i )
	   {
		*(shg+npel*(nsd2+1)*k+i) = *(xs)*(*(shl+npel*(nsd2+1)*k+i))+
			*(xs+2)*(*(shl+npel*(nsd2+1)*k+npel*1+i));
		*(shg+npel*(nsd2+1)*k+npel*1+i)=*(xs+1)*(*(shl+npel*(nsd2+1)*k+i))+
			*(xs+3)*(*(shl+npel*(nsd2+1)*k+npel*1+i));
		/*printf("%4d %16.4e %16.4e %16.4e\n", i, *(shg+npel*(nsd2+1)*k+i),
			*(shg+npel*(nsd2+1)*k+npel*1+i), 
			*(shg+npel*(nsd2+1)*k+npel*2+i));*/
	   }
	}
	return 1; 
}

