/*
    This utility function calculates the value of the electric field for each node
    on a 3 node triangle element.  This is done element by element where the node values
    use the already calculated edge values.  Each edge has a tangentinal and normal
    component which will be decomposed into X and Y global directions at each node.
    Because there may be multiple edges which may contain the node, averaging is done
    at the end based on the number of edges which have the node.  This is used
    by the electro-magnetic triangle element.

		Updated 12/16/09

    SLFCEM source file
    Version:  1.1
    Copyright (C) 2009  San Le 

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.
 
*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "treconst.h"
#include "trestruct.h"

#define SMALL          1.0e-12

extern int dof, numnp, numel, numed;
extern double shg[sosh], shg_node[sosh], shl[sosh], shg_E[sosh_E], shg_node_E[sosh_E],
	curl_E[num_int*epel], shl_node[sosh], shl_node2[sosh_node2], w[num_int], *Area0;

int matXT(double *, double *, double *, int, int, int);

int trshg( double *, int , int , double *, double *, double *);

int trshg_E( double *, int , int , double *, double *, double *, double *, double *,
	double *, double *);

int trEMnode_calculate(int *connect, int *edge_connect, int *node_edge_connect,
	double *coord, double *EMedge, double *EMnode, double *EMmag,
	double *node_counter)
{
	int i, i1, i2, j, k, dof_el[neqel], EMdof_el[neqel3], sdof_el[npel*nsd];;
	int check, counter, node, node0, node1, node2, node_el[npel];
	int matl_num, edge;
	double fdum, fdum1, fdum2, fdum3, fdum4;
	double EMedge_el[neqel3], EMnode_el[nsd];
	double vec_dum1[nsd], vec_dum2[nsd], vec_dum3[nsd];
	double coord_el_trans[npel*nsd], length[epel], lengthsq[epel];
	double Bedge[soB6];
	double X1, X2, X3, Y1, Y2, Y3;
	double det[num_int];
	double shg_sign[3];

/* Note below that I am still looping through the elements rather than the number of
   edges, "numed".  The reason is because I have to use the edge shape functions
   to get the normal electric field components to an edge.  These shape functions need the
   edge connectivity that is based on the elements.  Look at Figure 8.2 on page 234 of:

      Jin, Jianming, The Finite Element Method in Electromagnetics, John Wiley & Sons,
         New York, 1993.

   Basically, the directions of the electric field and edge numbering are:

                     
                   /
                  / /| /\
                 2 / | |
                / /  | 1
               / /---| 
              |/ -> 0
*/

	memset(EMnode,0,dof*sizeof(double));
	memset(EMmag,0,numnp*sizeof(double));

	/*printf("\n");*/
	for( k = 0; k < numel; ++k )
	{

/* Create the coord transpose vector and other variables for one element */

	    for( j = 0; j < npel; ++j )
	    {
		*(node_el+j) = node = *(connect+npel*k+j);

		*(sdof_el+nsd*j)=nsd*node;
		*(sdof_el+nsd*j+1)=nsd*node+1;

		*(coord_el_trans+j) = *(coord+*(sdof_el+nsd*j));
		*(coord_el_trans+npel*1+j) = *(coord+*(sdof_el+nsd*j+1));

		*(dof_el+ndof*j) = ndof*node;
		*(dof_el+ndof*j+1) = ndof*node+1;
	    }

/* Create edge_el for one element */

	    for( j = 0; j < epel; ++j )
	    {
		    edge = *(edge_connect+epel*k+j);
		    node1 = *(node_edge_connect+edge*nped);
		    node2 = *(node_edge_connect+edge*nped + 1);

		    *(EMdof_el+edof*j) = edof*edge;

/* Calculate the edge lengths */

		    fdum = *(coord+node2*nsd) - *(coord+node1*nsd);
		    fdum2 = *(coord+node2*nsd+1) - *(coord+node1*nsd+1);
		    *(lengthsq + j) = fdum*fdum + fdum2*fdum2;
		    *(length + j) = sqrt(*(lengthsq + j));

		    /*printf("%16.8e  %4d %4d\n", *(EMedge + edge), node1, node2);*/
	    }

/* Assembly of the global P matrix */

	    for( j = 0; j < neqel3; ++j )
	    {
		*(EMedge_el + j) = *(EMedge + *(EMdof_el+j));
		/*printf("%12.8e  %4d\n", *(EMedge_el + j), *(EMdof_el+j));*/
	    }

	    X1 = *(coord_el_trans);       Y1 = *(coord_el_trans + npel*1);
	    X2 = *(coord_el_trans + 1);   Y2 = *(coord_el_trans + npel*1 + 1);
	    X3 = *(coord_el_trans + 2);   Y3 = *(coord_el_trans + npel*1 + 2);

	    fdum = (X2 - X1)*(Y3 - Y1) - (X3 - X1)*(Y2 - Y1);

/* The variables below are needed to address the issue explained in:

   ~/slfcem-1.1/EDGE-ELEMENTS

*/
	    node0 = *(connect+npel*k);
	    node1 = *(connect+npel*k+1);
	    node2 = *(connect+npel*k+2);

	    shg_sign[0] = 1.0; shg_sign[1] = 1.0; shg_sign[2] = 1.0;

	    if(node0 > node1) shg_sign[0] = -1.0;
	    if(node1 > node2) shg_sign[1] = -1.0;
	    if(node2 > node0) shg_sign[2] = -1.0;

/* Assembly of the shg matrix for each nodal point */

	    check=trshg_E( det, k, 1, shl_node, shg_node, shg_node_E, curl_E,
		    coord_el_trans, length, shg_sign);
	    if(!check) printf( "Problems with trshg_E\n");

	    for( j = 0; j < num_int; ++j )
	    {
		*(Bedge)            = *(shg_node_E + epel*nsd2*j);
		*(Bedge + 1)        = *(shg_node_E + epel*nsd2*j + 1);
		*(Bedge + 2)        = *(shg_node_E + epel*nsd2*j + 2);
		*(Bedge+neqel3*1)   = *(shg_node_E + epel*nsd2*j + epel*1);
		*(Bedge+neqel3*1+1) = *(shg_node_E + epel*nsd2*j + epel*1 + 1);
		*(Bedge+neqel3*1+2) = *(shg_node_E + epel*nsd2*j + epel*1 + 2);

		check = matX(EMnode_el, Bedge, EMedge_el, nsd2, 1, neqel3);
		if(!check) printf( "Problems with matX \n");

/* Note that the EMnode_el is not like the U_el in my mechanics codes.  This EMnode_el[2]
   is the x and y component of node_el[j].   And an additional fact is that all
   intepolation is to a point, even if it edges are used as the interpolating quantity.

   The interpolation of EMnode is based on equation 8.15 on page 235 of Jin.

   The electric field magnitude, "EMmag", must be calculated as shown because
   at the location of the point charge, the different components will cancel
   each other out, but this is where the field is strongest.
*/
		*(EMnode + ndof*(*(node_el+j)))     += *(EMnode_el);
		*(EMnode + ndof*(*(node_el+j)) + 1) += *(EMnode_el + 1);

		fdum  = *(EMnode_el);
		fdum2 = *(EMnode_el+1);
		*(EMmag + *(node_el+j)) += sqrt(fdum*fdum + fdum2*fdum2);

		/*printf("\n wwwww %4d %4d\n", k, j);
		printf("%16.8e %16.8e %16.8e\n",
			*(Bedge), *(Bedge+1), *(Bedge+2));
		printf("%16.8e %16.8e %16.8e\n",
			*(Bedge+neqel3), *(Bedge+neqel3+1), *(Bedge+neqel3+2));*/

		/*printf("%4d %4d EMnode %16.8e %16.8e\n", k, j, *(EMnode_el), *(EMnode_el + 1));
		printf("%4d %4d Bedge %16.8e %16.8e %16.8e\n", k, j,
			*(Bedge), *(Bedge + 1), *(Bedge + 2));
		printf("%4d %4d Bedge %16.8e %16.8e %16.8e\n", k, j,
			*(Bedge+neqel3*1), *(Bedge+neqel3*1 + 1), *(Bedge+neqel3*1 + 2));*/
	    }

	}

/* Average all the components of EMnode depending based on its node count */

#if 1
	for( i = 0; i < numnp ; ++i )
	{
		/*printf("\nqqqq EMnode %4d %16.8e %16.8e %16.8e",
			i, *(node_counter+i), *(EMnode + ndof*i), *(EMnode + ndof*i+1));*/
		/**(EMnode + ndof*i)   /= *(node_counter+i);
		*(EMnode + ndof*i+1) /= *(node_counter+i);*/
		/*printf("\nqqqq EMnode %4d %16.8e %16.8e %16.8e",
			i, *(node_counter+i), *(EMnode + ndof*i), *(EMnode + ndof*i+1));*/
	}
#endif

	return 1;
}


int trEVnode_calculate(int *connect, double *coord, double *Volt, double *EMnode, double *EMmag,
	double *node_counter)
{
/*  The electric field is calculated at all four nodes of an element based on
    the gradient of the voltage potential.  Because there may be multiple
    elements which contain the node, averaging is done at the end based on
    the number of times the node appears.  This function is used when doing the
    electrostatic problem.

		Updated 12/18/09
*/
	int i, i1, i2, j, k, dof_el[neqel], Vdof_el[neqel3], sdof_el[npel*nsd];;
	int check, counter, node, node0, node1, node2, node3, node_el[npel];
	int matl_num, edge;
	double fdum, fdum1, fdum2, fdum3, fdum4;
	double gradB[soB6];
	double Volt_el[Vneqel3], EMnode_el[nsd];
	double coord_el_trans[npel*nsd];
	double det[num_int];
	double shg_sign[3];

	memset(EMnode,0,dof*sizeof(double));
	memset(EMmag,0,numnp*sizeof(double));

	for( k = 0; k < numel; ++k )
	{

/* Create the coord transpose vector and other variables for one element */

	    for( j = 0; j < npel; ++j )
	    {
		    *(node_el+j) = node = *(connect+npel*k+j);

		    *(sdof_el+nsd*j) = nsd*node;
		    *(sdof_el+nsd*j+1) = nsd*node+1;

		    *(coord_el_trans+j) = *(coord+*(sdof_el+nsd*j));
		    *(coord_el_trans+npel*1+j) = *(coord+*(sdof_el+nsd*j+1));

		    *(dof_el+ndof*j) = ndof*node;
		    *(dof_el+ndof*j+1) = ndof*node+1;

		    *(Vdof_el+Vndof*j) = Vndof*node;
	    }

	    for( j = 0; j < Vneqel3; ++j )
	    {
		*(Volt_el + j) = *(Volt + *(Vdof_el+j));
		/*printf("%12.8e  %4d\n", *(Volt_el + j), *(Vdof_el+j));*/
	    }

	    check=trshg( det, k, 1, shl_node, shg_node, coord_el_trans);
	    if(!check) printf( "Problems with qdshg\n");

	    /*printf("element %4d\n", k);*/
	    for( j = 0; j < npel; ++j )
	    {
		*(gradB)              = *(shg_node + npel*(nsd2+1)*j);
		*(gradB + 1)          = *(shg_node + npel*(nsd2+1)*j + 1);
		*(gradB + 2)          = *(shg_node + npel*(nsd2+1)*j + 2);

		*(gradB + npel*1)     = *(shg_node + npel*(nsd2+1)*j + npel*1);
		*(gradB + npel*1 + 1) = *(shg_node + npel*(nsd2+1)*j + npel*1 + 1);
		*(gradB + npel*1 + 2) = *(shg_node + npel*(nsd2+1)*j + npel*1 + 2);

		check = matX(EMnode_el, gradB, Volt_el, nsd2, 1, Vneqel3);
		if(!check) printf( "Problems with matX \n");

/* Note that the EMnode_el is not like the U_el in my mechanics codes.  This EMnode_el[2]
   is the x and y component of the electric field at node_el[j].

   The electric field is the negative of the gradient of the potential.

   The electric field magnitude, "EMmag", must be calculated as shown because
   at the location of the point charge, the different components will cancel
   each other out, but this is where the field is strongest.
*/
		*(EMnode + ndof*(*(node_el+j)))     -= *(EMnode_el);
		*(EMnode + ndof*(*(node_el+j)) + 1) -= *(EMnode_el + 1);

		fdum  = *(EMnode_el);
		fdum2 = *(EMnode_el+1);
		*(EMmag + *(node_el+j)) += sqrt(fdum*fdum + fdum2*fdum2);

		/*printf("%16.8e %16.8e %16.8e %16.8e\n",
			*(gradB), *(gradB + 1), *(gradB + 2), *(gradB + 3));
		printf("%16.8e %16.8e %16.8e %16.8e\n",
			*(gradB + npel*1), *(gradB + npel*1 + 1), *(gradB + npel*1 + 2), *(gradB + npel*1 + 3));
		printf("Volt_el %16.8e %16.8e %16.8e %16.8e\n",
			*(Volt_el), *(Volt_el + 1), *(Volt_el + 2), *(Volt_el + 3));*/

		/*printf("EMnode %4d %4d %16.8e %16.8e\n", k, *(node_el+j), *(EMnode_el), *(EMnode_el + 1));*/
	    }

	}

/* Average all the components of EMnode depending based on its node count */

	for( i = 0; i < numnp ; ++i )
	{
		*(EMnode + ndof*i)   /= *(node_counter+i);
		*(EMnode + ndof*i+1) /= *(node_counter+i);
		*(EMmag + i) /= *(node_counter+i);
	}

	return 1;
}

