/*
     These shape functions are for triangle elements which are used
     in electro-magnetic codes.

     SLFCEM source file
     Version:  1.1
     Copyright (C) 1999-2009  San Le

     The source code contained in this file is released under the
     terms of the GNU Library General Public License.
*/

#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "treconst.h"

#define  DEBUG         0

int dotX(double *,double *, double *, int );

int trshl( int integ_flag, double *shl, double *w)
{
/* 
     This subroutine calculates the local shape function derivatives for
     a triangle element at the gauss points.

     Because the derivatives of triangle elements are constant,
     this function is not really needed.  It is only here for testing
     purposes.

     It is a C translation of the subroutine QDCSHL from the book
     "The Finite Element Method" by Thomas Hughes, page 784.

     To see the shape functions for triangles, look on page 167.

     Also, look on page 13-3 to 13-6 in the "ANSYS User's Manual".  It
     discusses integration points for triangles.

     It should be noted that I have permutated the node sequences
     given by Hughes so that in terms of the triangle given in
     Figure 3.I.5 on page 167, node 3 in Hughes is node 1 in SLFCEM,
     node 2 is node 3, and node 1 goes to node 2.  This is because
     I want to be consistant with the tetrahedron.  You can read
     more about this change in teshl for the tetrahedron.

 ....  CALCULATE INTEGRATION-RULE WEIGHTS, SHAPE FUNCTIONS AND
       LOCAL DERIVATIVES FOR A EIGHT-NODE HEXAHEDRAL ELEMENT
       r, s = LOCAL ELEMENT COORDINATES along triangle faces
       *(shl+npel*(nsd2+1)*k+i) = LOCAL ("r") DERIVATIVE OF SHAPE FUNCTION
       *(shl+npel*(nsd2+1)*k+npel*1+i) = LOCAL ("s") DERIVATIVE OF SHAPE FUNCTION
       *(shl+npel*(nsd2+1)*k+npel*2+i) = LOCAL SHAPE FUNCTION
       *(w+k)    = INTEGRATION-RULE WEIGHT
       i       = LOCAL NODE NUMBER
       k       = INTEGRATION POINT
       num_int = NUMBER OF INTEGRATION POINTS, EQ. 1 OR 3

                        Updated 11/26/09
*/

	double ra[3], sa[3];
	double temps,tempr,r,s,fdum;
	int i,j,k;

	if( integ_flag == 0)
	{

/* Set ra, sa, ta for shl_node calculation */

		*(ra) = 0.0;     *(sa) = 0.0;
		*(ra + 1) = 1.0; *(sa + 1) = 0.0;
		*(ra + 2) = 0.0; *(sa + 2) = 1.0;
	}

	if( integ_flag == 1)
	{

/* Set ra, sa, ta for 3 point gauss calculation */

		*(ra) = pt1667;     *(sa) = pt1667;
		*(ra + 1) = pt6667; *(sa + 1) = pt1667;
		*(ra + 2) = pt1667; *(sa + 2) = pt6667;

		*(w)=pt3333;
		*(w + 1)=pt3333;
		*(w + 2)=pt3333;
	}

	if( integ_flag == 2)
	{

/* Set ra, sa, ta for 1 point gauss calculation.  For 1 point gauss, there
   is only one "ra", "sa", and "w" used so I set w[1] = w[2] = 0.0. */

		*(ra) = 1.0/3.0;     *(sa) = 1.0/3.0;
		*(ra + 1) = 1.0/3.0; *(sa + 1) = 1.0/3.0;
		*(ra + 2) = 1.0/3.0; *(sa + 2) = 1.0/3.0;

		*(w) = 1.0;
		*(w + 1) = 0.0;
		*(w + 2) = 0.0;
	}

	for( k = 0; k < num_int; ++k )
	{
/* calculating the weights and local dN/ds,dr matrix */

		r=*(ra+k);
		s=*(sa+k);
		fdum = (1.0 - r - s);

/* dN/dr */
		*(shl+npel*(nsd2+1)*k)=   -1.0;         /* node 0 */
		*(shl+npel*(nsd2+1)*k+1)=  1.0;         /* node 1 */
		*(shl+npel*(nsd2+1)*k+2)=  0.0;         /* node 2 */

/* dN/ds */
		*(shl+npel*(nsd2+1)*k+npel*1)=  -1.0;   /* node 0 */
		*(shl+npel*(nsd2+1)*k+npel*1+1)= 0.0;   /* node 1 */
		*(shl+npel*(nsd2+1)*k+npel*1+2)= 1.0;   /* node 2 */

/* N */
		*(shl+npel*(nsd2+1)*k+npel*2)=   fdum;  /* node 0 */
		*(shl+npel*(nsd2+1)*k+npel*2+1)= r;     /* node 1 */
		*(shl+npel*(nsd2+1)*k+npel*2+2)= s;     /* node 2 */

		/*printf("\n");*/

		/*for( i = 0; i < nsd2+1; ++i )
		{
		    for( j = 0; j < npel; ++j )
		    {
			printf(" %14.6e",*(shl+npel*(nsd2+1)*k+npel*i + j));
		    }
		    printf("\n");
		}
		printf("\n");*/
	}
	return 1;
}

int trshg_E( double *det, int el, int integ_flag, double *shl, double *shg,
	double *shg_E, double *curl_E, double *xl, double *length, double *shg_sign)
{
/*
     This subroutine calculates the global shape function derivatives for
     a quad element at the gauss points.

     Because the derivatives of triangle elements are constant, 
     this function is not really needed.  It is only here for testing
     purposes.

     It is a C translation of the subroutine QDCSHG from the book
     "The Finite Element Method" by Thomas Hughes, page 783.

 ....  CALCULATE GLOBAL DERIVATIVES OF SHAPE FUNCTIONS AND
       JACOBIAN DETERMINANTS FOR A FOUR-NODE QUADRALATERAL ELEMENT

       *(xl+j+npel*i) = GLOBAL COORDINATES(I LOOPS OVER X,Y)
       *(det)  = JACOBIAN DETERMINANT
       *(shl+npel*(nsd2+1)*k+i) = LOCAL ("r") DERIVATIVE OF SHAPE FUNCTION
       *(shl+npel*(nsd2+1)*k+npel*1+i) = LOCAL ("s") DERIVATIVE OF SHAPE FUNCTION
       *(shl+npel*(nsd2+1)*k+npel*2+i) = LOCAL SHAPE FUNCTION
       *(shg+npel*(nsd2+1)*k+i) = X-DERIVATIVE OF SHAPE FUNCTION
       *(shg+npel*(nsd2+1)*k+npel*1+i) = Y-DERIVATIVE OF SHAPE FUNCTION
       *(shg+npel*(nsd2+1)*k+npel*2+i) = shl(npel*(nsd2+1)*k+npel*2+i)
       *(xs+2*i+j) = JACOBIAN MATRIX
          i    = LOCAL NODE NUMBER OR GLOBAL COORDINATE NUMBER
          j    = GLOBAL COORDINATE NUMBER
          k    = INTEGRATION-POINT NUMBER
       num_int    = NUMBER OF INTEGRATION POINTS, EQ.3

          i    = LOCAL EDGE NUMBER
       *(shg_E+epel*nsd2*k+i) = X COMPONENT OF SHAPE FUNCTION
       *(shg_E+epel*nsd2*k+epel*1+i) = Y COMPONENT OF SHAPE FUNCTION
       *(curl_E + i) = CURL OF SHAPE FUNCTION

                        Updated 12/16/09
*/

	double xs[4],temp;
	int check,i,j,k;
	double X1, X2, X3, Y1, Y2, Y3, XX, YY, Area, fdum;
	double a1, a2, a3, b1, b2, b3, c1, c2, c3;
	double L1, L2, L3, gradL1x, gradL1y, gradL2x, gradL2y, gradL3x, gradL3y;

	memcpy(shg,shl,sosh*sizeof(double));

	X1 = *(xl);
	X2 = *(xl + 1);
	X3 = *(xl + 2);

	Y1 = *(xl + npel*1);
	Y2 = *(xl + npel*1 + 1);
	Y3 = *(xl + npel*1 + 2);

/* The formulas below are from page 80 of Jin, below equation 4.24. */

	a1 = X2*Y3 - Y2*X3;  b1 = Y2 - Y3;  c1 = X3 - X2;
	a2 = X3*Y1 - Y3*X1;  b2 = Y3 - Y1;  c2 = X1 - X3;
	a3 = X1*Y2 - Y1*X2;  b3 = Y1 - Y2;  c3 = X2 - X1;

	Area = (X2 - X1)*(Y3 - Y1) - (X3 - X1)*(Y2 - Y1);
	Area *= 0.5;

/* The jacobian dx/dc is calculated below */

#if 0
	for( j = 0; j < nsd2; ++j )
	{
		for( i = 0; i < nsd2; ++i )
		{
		   check=dotX((xs+nsd2*i+j),(shg+npel*j), (xl+npel*i),npel);
		}
	}
#endif

#if 1
	*(xs) =   - *(xl) + *(xl + 1);
	*(xs+1) = - *(xl) + *(xl + 2);
	*(xs+2) = - *(xl + 3) + *(xl + 4);
	*(xs+3) = - *(xl + 3) + *(xl + 5);
#endif

	*(det)=*(xs)*(*(xs+3))-*(xs+2)*(*(xs+1));
	/*printf(" %9.5f %9.5f\n %9.5f %9.5f\n",
		*(xs),*(xs+1),*(xs+nsd2*1),*(xs+nsd2*1+1));
	printf("%d %f\n", k, *(det));*/

	if(*(det) <= 0.0 )
	{
		printf("the element (%d) is inverted; det:%f\n", el,*(det));
		return 0;
	}

/* The inverse of the jacobian, dc/dx, is calculated below */

	temp=*(xs);
	*(xs)=*(xs+3)/(*(det));
	*(xs+1)*=-1./(*(det));
	*(xs+2)*=-1./(*(det));
	*(xs+3)=temp/(*(det));

	if(integ_flag == 1)
	{
/* For 3 point gauss */
	    for( k = 0; k < num_int; ++k )
	    {
		for( i = 0; i < npel; ++i )
		{
		   *(shg+npel*(nsd2+1)*k+i) = *(xs)*(*(shl+npel*(nsd2+1)*k+i))+
			*(xs+2)*(*(shl+npel*(nsd2+1)*k+npel*1+i));
		   *(shg+npel*(nsd2+1)*k+npel*1+i)=*(xs+1)*(*(shl+npel*(nsd2+1)*k+i))+
			*(xs+3)*(*(shl+npel*(nsd2+1)*k+npel*1+i));
		   /*printf("%d %f %f %f\n", i, *(shg+npel*(nsd2+1)*k+i),
			*(shg+npel*(nsd2+1)*k+npel*1+i),
			*(shg+npel*(nsd2+1)*k+npel*2+i));*/
		}
	    }
	}

	if(integ_flag == 2)
	{
/* For 1 point gauss */
	   for( i = 0; i < npel; ++i )
	   {
		*(shg+i) = *(xs)*(*(shl+i))+ *(xs+2)*(*(shl+npel*1+i));
		*(shg+npel*1+i)=*(xs+1)*(*(shl+i))+ *(xs+3)*(*(shl+npel*1+i));
		/*printf("%d %f %f %f\n", i, *(shg+i), *(shg+npel*1+i), *(shg+npel*2+i));*/
	   }
	}

/*
     Calculate the local edge shape function derivatives for a triangle element.

     Unfortunately, the only way to understand how it all works is to have
     all the notes I have taken as well as Jin's book.  The relevant equations
     are the equations below 4.24 on page 80 as mentioned above for a1, a2, a3,
     b1, b2, b3, c1, c2, c3.  For the equations of the edge based shape functions,
     given by "shg_E" below, look at equation 8.12-8.14 on page 235.  For the L's
     you see in 8.12-8.14, which I do not define, look at equations 4.158-4.160
     on page 126-127.  The curl of the shape functions, which are given by
     "curl_E" below, are not defined in the book.

     Note that the edge based interpolating shape functions below are different
     from the node based above.  This edge interpolation is not encountered
     in my other codes which is why it is expressed so ackwardly.

     Checked shg_E and curlBedge and they are correct.  The 1/4 is already
     in shl.

     I am using 3 point integration for the triangle.

*/

	fdum = 4.0*Area*Area;

	if(integ_flag == 1)
	{
/* For 3 point gauss */
	    /*printf("\n");*/
	    for( k = 0; k < num_int; ++k )
	    {
                XX = *(shg+npel*(nsd2+1)*k+npel*2)*X1 + *(shg+npel*(nsd2+1)*k+npel*2 + 1)*X2 +
                    *(shg+npel*(nsd2+1)*k+npel*2 + 2)*X3;
                YY = *(shg+npel*(nsd2+1)*k+npel*2)*Y1 + *(shg+npel*(nsd2+1)*k+npel*2 + 1)*Y2 +
                    *(shg+npel*(nsd2+1)*k+npel*2 + 2)*Y3;

#if !DEBUG 
                L1 = (a1 + b1*XX + c1*YY)/Area/2.0;
                L2 = (a2 + b2*XX + c2*YY)/Area/2.0;
                L3 = (a3 + b3*XX + c3*YY)/Area/2.0;

                gradL1x = b1/Area/2.0; gradL1y = c1/Area/2.0;
                gradL2x = b2/Area/2.0; gradL2y = c2/Area/2.0;
                gradL3x = b3/Area/2.0; gradL3y = c3/Area/2.0;

                /*printf("%16.8e %16.8e %16.8e %16.8e %16.8e %16.8e\n", X1, X2, X3, Y1, Y2, Y3);
                printf("%16.8e %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e\n",
                    a1, a2, a3, b1, b2, b3, c1, c2, c3);*/
	        /*printf("\n");
                printf("%16.8e %16.8e %16.8e\n", L1, L2, L3);
                printf("%16.8e %16.8e %16.8e\n", gradL1x, gradL2x, gradL3x);
                printf("%16.8e %16.8e %16.8e\n", gradL1y, gradL2y, gradL3y);*/

                *(shg_E+epel*nsd2*k)   =
			*(shg_sign)*(*(length))*(L1*gradL2x - L2*gradL1x);
                *(shg_E+epel*nsd2*k+1) =
			*(shg_sign+1)*(*(length+1))*(L2*gradL3x - L3*gradL2x);
                *(shg_E+epel*nsd2*k+2) =
			*(shg_sign+2)*(*(length+2))*(L3*gradL1x - L1*gradL3x);

                *(shg_E+epel*nsd2*k+epel*1)   =
			*(shg_sign)*(*(length))*(L1*gradL2y - L2*gradL1y);
                *(shg_E+epel*nsd2*k+epel*1+1) =
			*(shg_sign+1)*(*(length+1))*(L2*gradL3y - L3*gradL2y);
                *(shg_E+epel*nsd2*k+epel*1+2) =
			*(shg_sign+2)*(*(length+2))*(L3*gradL1y - L1*gradL3y);

                /*printf("%16.8e %16.8e %16.8e\n",
			*(shg_E+epel*nsd2*k),
			*(shg_E+epel*nsd2*k+1),
			*(shg_E+epel*nsd2*k+2));
                printf("%16.8e %16.8e %16.8e\n", 
			*(shg_E+epel*nsd2*k+epel*1),
			*(shg_E+epel*nsd2*k+epel*1+1),
			*(shg_E+epel*nsd2*k+epel*1+2));*/
                /*printf("%16.8e %16.8e %16.8e\n",
			(L1*gradL2x - L2*gradL1x), (L2*gradL3x - L3*gradL2x), (L3*gradL1x - L1*gradL3x));*/
                /*printf("%4d %4d %16.8e %16.8e %16.8e\n", el, k, 
			(L1*gradL2y - L2*gradL1y), (L2*gradL3y - L3*gradL2y), (L3*gradL1y - L1*gradL3y));*/
                /*printf("%16.8e %16.8e %16.8e\n", *(shg_sign), *(shg_sign+1), *(shg_sign+2));*/
#endif
#if DEBUG 
                *(shg_E+epel*nsd2*k) =
                    *(shg_sign)*(*(length))*((a1*b2 - a2*b1) + (b2*c1 - b1*c2)*YY)/fdum;
                *(shg_E+epel*nsd2*k+1) =
                    *(shg_sign+1)*(*(length + 1))*((a2*b3 - a3*b2) + (b3*c2 - b2*c3)*YY)/fdum;
                *(shg_E+epel*nsd2*k+2) =
                    *(shg_sign+2)*(*(length + 2))*((a3*b1 - a1*b3) + (b1*c3 - b3*c1)*YY)/fdum;

                /*printf("%16.8e %16.8e %16.8e\n",
			((a1*b2 - a2*b1) + (b2*c1 - b1*c2)*YY)/fdum, ((a2*b3 - a3*b2) + (b3*c2 - b2*c3)*YY)/fdum,
			((a3*b1 - a1*b3) + (b1*c3 - b3*c1)*YY)/fdum);*/
                /*printf("%16.8e %16.8e %16.8e\n", *(shg_sign), *(shg_sign+1), *(shg_sign+2));*/

                *(shg_E+epel*nsd2*k+epel*1) =
                    *(shg_sign)*(*(length))*((a1*c2 - a2*c1) + (b1*c2 - b2*c1)*XX)/fdum;
                *(shg_E+epel*nsd2*k+epel*1+1) =
                    *(shg_sign+1)*(*(length + 1))*((a2*c3 - a3*c2) + (b2*c3 - b3*c2)*XX)/fdum;
                *(shg_E+epel*nsd2*k+epel*1+2) =
                    *(shg_sign+2)*(*(length + 2))*((a3*c1 - a1*c3) + (b3*c1 - b1*c3)*XX)/fdum;

                /*printf("%4d %4d %16.8e %16.8e %16.8e\n", el, k, 
			((a1*c2 - a2*c1) + (b1*c2 - b2*c1)*XX)/fdum, ((a2*c3 - a3*c2) + (b2*c3 - b3*c2)*XX)/fdum,
			((a3*c1 - a1*c3) + (b3*c1 - b1*c3)*XX)/fdum);*/
                /*printf("%16.8e %16.8e %16.8e\n",
			*(shg_E+epel*nsd2*k),
			*(shg_E+epel*nsd2*k+1),
			*(shg_E+epel*nsd2*k+2));
                printf("%16.8e %16.8e %16.8e\n", 
			*(shg_E+epel*nsd2*k+epel*1),
			*(shg_E+epel*nsd2*k+epel*1+1),
			*(shg_E+epel*nsd2*k+epel*1+2));*/
#endif
	    }
	}

	if(integ_flag == 2)
	{
/* For 1 point gauss */

		k = 0;
                XX = *(shg+npel*2)*X1 + *(shg+npel*2 + 1)*X2 + *(shg+npel*2 + 2)*X3;
                YY = *(shg+npel*2)*Y1 + *(shg+npel*2 + 1)*Y2 + *(shg+npel*2 + 2)*Y3;
#if !DEBUG 
                L1 = (a1 + b1*XX + c1*YY)/Area/2.0;
                L2 = (a2 + b2*XX + c2*YY)/Area/2.0;
                L3 = (a3 + b3*XX + c3*YY)/Area/2.0;

                gradL1x = b1/Area/2.0; gradL1y = c1/Area/2.0;
                gradL2x = b2/Area/2.0; gradL2y = c2/Area/2.0;
                gradL3x = b3/Area/2.0; gradL3y = c3/Area/2.0;

                /*printf("%16.8e %16.8e %16.8e %16.8e %16.8e %16.8e\n", X1, X2, X3, Y1, Y2, Y3);
                printf("%16.8e %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e\n",
                    a1, a2, a3, b1, b2, b3, c1, c2, c3);*/
	        /*printf("\n");
                printf("%16.8e %16.8e %16.8e\n", L1, L2, L3);
                printf("%16.8e %16.8e %16.8e\n", gradL1x, gradL2x, gradL3x);
                printf("%16.8e %16.8e %16.8e\n", gradL1y, gradL2y, gradL3y);*/

                *(shg_E)   = *(shg_sign)*(*(length))*(L1*gradL2x - L2*gradL1x);
                *(shg_E+1) = *(shg_sign+1)*(*(length+1))*(L2*gradL3x - L3*gradL2x);
                *(shg_E+2) = *(shg_sign+2)*(*(length+2))*(L3*gradL1x - L1*gradL3x);

                *(shg_E+epel*1)   = *(shg_sign)*(*(length))*(L1*gradL2y - L2*gradL1y);
                *(shg_E+epel*1+1) = *(shg_sign+1)*(*(length+1))*(L2*gradL3y - L3*gradL2y);
                *(shg_E+epel*1+2) = *(shg_sign+2)*(*(length+2))*(L3*gradL1y - L1*gradL3y);

                /*printf("%16.8e %16.8e %16.8e\n", *(shg_E), *(shg_E+1), *(shg_E+2));
                printf("%16.8e %16.8e %16.8e\n", *(shg_E+epel*1), *(shg_E+epel*1+1), *(shg_E+epel*1+2));*/
                /*printf("%16.8e %16.8e %16.8e\n",
			(L1*gradL2x - L2*gradL1x), (L2*gradL3x - L3*gradL2x), (L3*gradL1x - L1*gradL3x));*/
                /*printf("%4d %4d %16.8e %16.8e %16.8e\n", el, k, 
			(L1*gradL2y - L2*gradL1y), (L2*gradL3y - L3*gradL2y), (L3*gradL1y - L1*gradL3y));*/
                /*printf("%16.8e %16.8e %16.8e\n", *(shg_sign), *(shg_sign+1), *(shg_sign+2));*/
#endif
#if DEBUG 
                *(shg_E) =
                    *(shg_sign)*(*(length))*((a1*b2 - a2*b1) + (b2*c1 - b1*c2)*YY)/fdum;
                *(shg_E+1) =
                    *(shg_sign+1)*(*(length + 1))*((a2*b3 - a3*b2) + (b3*c2 - b2*c3)*YY)/fdum;
                *(shg_E+2) =
                    *(shg_sign+2)*(*(length + 2))*((a3*b1 - a1*b3) + (b1*c3 - b3*c1)*YY)/fdum;

                /*printf("%16.8e %16.8e %16.8e\n",
			((a1*b2 - a2*b1) + (b2*c1 - b1*c2)*YY)/fdum, ((a2*b3 - a3*b2) + (b3*c2 - b2*c3)*YY)/fdum,
			((a3*b1 - a1*b3) + (b1*c3 - b3*c1)*YY)/fdum);*/
                /*printf("%16.8e %16.8e %16.8e\n", *(shg_sign), *(shg_sign+1), *(shg_sign+2));*/

                *(shg_E+epel*1) =
                    *(shg_sign)*(*(length))*((a1*c2 - a2*c1) + (b1*c2 - b2*c1)*XX)/fdum;
                *(shg_E+epel*1+1) =
                    *(shg_sign+1)*(*(length + 1))*((a2*c3 - a3*c2) + (b2*c3 - b3*c2)*XX)/fdum;
                *(shg_E+epel*1+2) =
                    *(shg_sign+2)*(*(length + 2))*((a3*c1 - a1*c3) + (b3*c1 - b1*c3)*XX)/fdum;

                /*printf("%4d %4d %16.8e %16.8e %16.8e\n", el, k, 
			((a1*c2 - a2*c1) + (b1*c2 - b2*c1)*XX)/fdum, ((a2*c3 - a3*c2) + (b2*c3 - b3*c2)*XX)/fdum,
			((a3*c1 - a1*c3) + (b3*c1 - b1*c3)*XX)/fdum);*/
                /*printf("%16.8e %16.8e %16.8e\n", *(shg_E), *(shg_E+1), *(shg_E+2));
                printf("%16.8e %16.8e %16.8e\n", *(shg_E+epel*1), *(shg_E+epel*1+1), *(shg_E+epel*1+2));*/
#endif
	}

/* Calculate the curl of the shape functions above */

	fdum = 2.0*Area*Area;
	*(curl_E) =
		*(shg_sign)*(*(length))*(b1*c2 - b2*c1)/fdum;
	*(curl_E + 1) =
		*(shg_sign+1)*(*(length + 1))*(b2*c3 - b3*c2)/fdum;
	*(curl_E + 2) =
		*(shg_sign+2)*(*(length + 2))*(b3*c1 - b1*c3)/fdum;

	/*printf("\n");*/



	return 1;
}

int trshg( double *det, int el, int integ_flag, double *shl, double *shg, double *xl)
{
/*
     This subroutine calculates the global shape function derivatives for
     a quad element at the gauss points.

     Because the derivatives of triangle elements are constant, 
     this function is not really needed.  It is only here for testing
     purposes.

     It is a C translation of the subroutine QDCSHG from the book
     "The Finite Element Method" by Thomas Hughes, page 783.

 ....  CALCULATE GLOBAL DERIVATIVES OF SHAPE FUNCTIONS AND
       JACOBIAN DETERMINANTS FOR A FOUR-NODE QUADRALATERAL ELEMENT

       *(xl+j+npel*i) = GLOBAL COORDINATES(I LOOPS OVER X,Y)
       *(det)  = JACOBIAN DETERMINANT
       *(shl+npel*(nsd2+1)*k+i) = LOCAL ("r") DERIVATIVE OF SHAPE FUNCTION
       *(shl+npel*(nsd2+1)*k+npel*1+i) = LOCAL ("s") DERIVATIVE OF SHAPE FUNCTION
       *(shl+npel*(nsd2+1)*k+npel*2+i) = LOCAL SHAPE FUNCTION
       *(shg+npel*(nsd2+1)*k+i) = X-DERIVATIVE OF SHAPE FUNCTION
       *(shg+npel*(nsd2+1)*k+npel*1+i) = Y-DERIVATIVE OF SHAPE FUNCTION
       *(shg+npel*(nsd2+1)*k+npel*2+i) = shl(npel*(nsd2+1)*k+npel*2+i)
       *(xs+2*i+j) = JACOBIAN MATRIX
          i    = LOCAL NODE NUMBER OR GLOBAL COORDINATE NUMBER
          j    = GLOBAL COORDINATE NUMBER
          k    = INTEGRATION-POINT NUMBER
       num_int    = NUMBER OF INTEGRATION POINTS, EQ.3

                        Updated 8/6/06
*/

	double xs[4],temp;
	int check,i,j,k;

	memcpy(shg,shl,sosh*sizeof(double));

/* The jacobian dx/dc is calculated below */

#if 0
	for( j = 0; j < nsd2; ++j )
	{
		for( i = 0; i < nsd2; ++i )
		{
		   check=dotX((xs+nsd2*i+j),(shg+npel*j), (xl+npel*i),npel);
		}
	}
#endif

#if 1
	*(xs) =   - *(xl) + *(xl + 1);
	*(xs+1) = - *(xl) + *(xl + 2);
	*(xs+2) = - *(xl + 3) + *(xl + 4);
	*(xs+3) = - *(xl + 3) + *(xl + 5);
#endif

	*(det)=*(xs)*(*(xs+3))-*(xs+2)*(*(xs+1));
	/*printf(" %9.5f %9.5f\n %9.5f %9.5f\n",
		*(xs),*(xs+1),*(xs+nsd2*1),*(xs+nsd2*1+1));
	printf("%d %f\n", k, *(det));*/

	if(*(det) <= 0.0 )
	{
		printf("the element (%d) is inverted; det:%f\n", el,*(det));
		return 0;
	}

/* The inverse of the jacobian, dc/dx, is calculated below */

	temp=*(xs);
	*(xs)=*(xs+3)/(*(det));
	*(xs+1)*=-1./(*(det));
	*(xs+2)*=-1./(*(det));
	*(xs+3)=temp/(*(det));

	if(integ_flag == 1)
	{
/* For 3 point gauss */
	    for( k = 0; k < num_int; ++k )
	    {
		for( i = 0; i < npel; ++i )
		{
		   *(shg+npel*(nsd2+1)*k+i) = *(xs)*(*(shl+npel*(nsd2+1)*k+i))+
			*(xs+2)*(*(shl+npel*(nsd2+1)*k+npel*1+i));
		   *(shg+npel*(nsd2+1)*k+npel*1+i)=*(xs+1)*(*(shl+npel*(nsd2+1)*k+i))+
			*(xs+3)*(*(shl+npel*(nsd2+1)*k+npel*1+i));
		   /*printf("%d %f %f %f\n", i, *(shg+npel*(nsd2+1)*k+i),
			*(shg+npel*(nsd2+1)*k+npel*1+i),
			*(shg+npel*(nsd2+1)*k+npel*2+i));*/
		}
	    }
	}

	if(integ_flag == 2)
	{
/* For 1 point gauss */
	   for( i = 0; i < npel; ++i )
	   {
		*(shg+i) = *(xs)*(*(shl+i))+ *(xs+2)*(*(shl+npel*1+i));
		*(shg+npel*1+i)=*(xs+1)*(*(shl+i))+ *(xs+3)*(*(shl+npel*1+i));
		/*printf("%d %f %f %f\n", i, *(shg+i), *(shg+npel*1+i), *(shg+npel*2+i));*/
	   }
	}


	return 1;
}


